#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <Errors.h>
#include <Files.h>
#include <Memory.h>
#include <Events.h>
#include <TextUtils.h>
#include <StandardFile.h>
#include <Notification.h>
#include <dialogs.h>
#include <SIOUX.h>
#include <SIOUXGlobals.h>
#include <console.h>

#define TRIMSTR(nam) {\
  char *p = nam; \
  if (*p == '"') memcpy(p, p+1, strlen(p)-1); \
  while (*p) { \
    if (*p == 13 || *p == 10 || *p == '"') \
    *p = 0; \
    p++; \
  } \
}
 
#define FALSE 0
#define TRUE 1
#define MESH_HEADER_EDITOR_DATA_SIZE 16
#define SKIP_SIZE 64

enum // next mesh array indexes
{
        _next_mesh_success,
        _next_mesh_failure,
        NUMBER_OF_NEXT_MESH_TAGS
};

enum // cutscenes array indexes
{
        _cutscene_pregame,
        _cutscene_postgame_success,
        _cutscene_postgame_failure,
        NUMBER_OF_CUTSCENE_MOVIE_TAGS
};

enum // screens array indexes (these are just collections)
{
        _screen_loss,
        _screen_success,
        _screen_failure,
        NUMBER_OF_SCREEN_COLLECTION_TAGS
};

enum // storyline array indexes
{
  _storyline_pregame,
  _storyline_unused0,
  _storyline_unused1,
  _storyline_unused2,
  NUMBER_OF_STORYLINE_STRING_TAGS
};

typedef int file_tag;
typedef unsigned short word;
typedef short short_fixed;
typedef short short_world_distance;
typedef long fixed;
typedef long world_distance;
typedef unsigned char byte;

struct rectangle2d
{
  short top, left, bottom, right;
};
typedef struct rectangle2d rectangle2d;


struct rgb_color
{
  word red, green, blue;
  word flags;
};


struct m2_mesh_header
{
  file_tag landscape_collection_tag;
  file_tag media_tag;
  
  short submesh_width, submesh_height;
  long mesh_offset, mesh_size;
  void *mesh;

  long data_offset, data_size;
  void *data;
  
  long marker_palette_entries, marker_palette_offset, marker_palette_size; // array of marker_data structures
  void *marker_palette; // combined with marker list in scenario_header
  
  long marker_count, markers_offset, markers_size; // array of marker_data structures
  void *markers;
  
  file_tag mesh_lighting_tag; // shadow intensity, ambient light, etc.
  file_tag connector_tag; // for fences
  
  unsigned long flags;
  
  file_tag particle_system_tag;

  long team_count;
  
  short_fixed dark_fraction;
  short_fixed light_fraction;
  struct rgb_color dark_color;
  struct rgb_color light_color;
  fixed transition_point;

  short_world_distance ceiling_height;
  short unused1;

  // this is not a rectangle; these are the buffers on each edge of the map, in cells
  rectangle2d edge_of_mesh_buffer_zones;
  
  file_tag global_ambient_sound_tag;

  long map_action_count, map_actions_offset, map_action_buffer_size;

  file_tag map_description_string_list_tag;
  file_tag postgame_collection_tag;
  file_tag pregame_collection_tag;
  file_tag overhead_map_collection_tag;
  file_tag next_mesh_tags[NUMBER_OF_NEXT_MESH_TAGS];
  file_tag cutscene_movie_tags[NUMBER_OF_CUTSCENE_MOVIE_TAGS];
  file_tag storyline_string_tags[NUMBER_OF_STORYLINE_STRING_TAGS];
  
  long media_coverage_region_offset, media_coverage_region_size;
  short *media_coverage_region;
  
  long mesh_LOD_data_offset, mesh_LOD_data_size;
  word *mesh_LOD_data;

  struct rgb_color global_tint_color;
  short_fixed global_tint_fraction;
  word pad;
  
  file_tag wind_tag;

  file_tag screen_collection_tags[NUMBER_OF_SCREEN_COLLECTION_TAGS];
  
  struct rgb_color blood_color;

  file_tag picture_caption_string_list_tag;
  file_tag narration_sound_tag;
  file_tag win_ambient_sound_tag;
  file_tag loss_ambient_sound_tag;

  // For Creative's Environmental Audio Extensions
  unsigned long reverb_environment;
  float reverb_volume;
  float reverb_decay_time;
  float reverb_damping;

  long connector_count, connectors_offset, connectors_size; // array of connectors (i.e. fences)
  struct saved_connector_data *connectors;

  char cutscene_names[NUMBER_OF_CUTSCENE_MOVIE_TAGS][64];

  file_tag hints_string_list_tag;

  struct rgb_color fog_color;
  float fog_density;

  file_tag difficulty_level_override_string_list_tag;

  short unused[241];
  
  short connector_type;
  short map_description_string_index;
  short overhead_map_collection_index;
  short landscape_collection_index;
  short global_ambient_sound_index;
  short media_type;
  short hints_string_list_index;
  
  byte editor_data[MESH_HEADER_EDITOR_DATA_SIZE];
};

struct tfl_mesh_header
{
  file_tag landscape_collection_tag;
  file_tag media_tag;
  
  short submesh_width, submesh_height;
  long mesh_offset, mesh_size;
  void *mesh;

  long data_offset, data_size;
  void *data;
  
  long marker_palette_entries, marker_palette_offset, marker_palette_size; // array of marker_data structures
  void *marker_palette; // combined with marker list in scenario_header
  
  long marker_count, markers_offset, markers_size; // array of marker_data structures
  void *markers;
  
  file_tag mesh_lighting_tag; // shadow intensity, ambient light, etc.
  file_tag connector_tag; // for fences
  
  unsigned long flags;
  
  file_tag particle_system_tag;

  long team_count;
  
  short_fixed dark_fraction;
  short_fixed light_fraction;
  struct rgb_color dark_color;
  struct rgb_color light_color;
  fixed transition_point;

  short_world_distance ceiling_height;
  short unused1;

  // this is not a rectangle; these are the buffers on each edge of the map, in cells
  rectangle2d edge_of_mesh_buffer_zones;
  
  file_tag global_ambient_sound_tag;

  long map_action_count, map_actions_offset, map_action_buffer_size;

  file_tag map_description_string_list_tag;
  file_tag postgame_collection_tag;
  file_tag pregame_collection_tag;
  file_tag overhead_map_collection_tag;
  file_tag next_mesh_tags[NUMBER_OF_NEXT_MESH_TAGS];
  file_tag cutscene_movie_tags[NUMBER_OF_CUTSCENE_MOVIE_TAGS];
  file_tag storyline_string_tags[NUMBER_OF_STORYLINE_STRING_TAGS];
  
  long media_coverage_region_offset, media_coverage_region_size;
  short *media_coverage_region;
  
  long mesh_LOD_data_offset, mesh_LOD_data_size;
  word *mesh_LOD_data;

  struct rgb_color global_tint_color;
  short_fixed global_tint_fraction;
  word pad;
  
  file_tag wind_tag;

  file_tag screen_collection_tags[NUMBER_OF_SCREEN_COLLECTION_TAGS];
  
  struct rgb_color blood_color;

  file_tag picture_caption_string_list_tag;
  file_tag narration_sound_tag;
  file_tag win_ambient_sound_tag;
  file_tag loss_ambient_sound_tag;

  // For Creative's Environmental Audio Extensions
  unsigned long reverb_environment;
  float reverb_volume;
  float reverb_decay_time;
  float reverb_damping;

  long connector_count, connectors_offset, connectors_size; // array of connectors (i.e. fences)
  struct saved_connector_data *connectors;

  char cutscene_names[NUMBER_OF_CUTSCENE_MOVIE_TAGS][64];

  //file_tag hints_string_list_tag;

  struct rgb_color fog_color;
  float fog_density;

  file_tag difficulty_level_override_string_list_tag;

  short unused[241];
  
  short connector_type;
  short map_description_string_index;
  short overhead_map_collection_index;
  short landscape_collection_index;
  short global_ambient_sound_index;
  short media_type;
  //short hints_string_list_index;
  
  byte editor_data[MESH_HEADER_EDITOR_DATA_SIZE];
};

enum
{
  _marker_team,
  _marker_scenery,
  _marker_unused1,
  _marker_monster,
  _marker_unused2,
  _marker_sound_source,
  _marker_model,
  _marker_unused3,
  _marker_unused4,
  _marker_projectile,
  _marker_local_projectile_group
};

struct marker_data
{
  int flags;
  short type;
  short palette;
  short identifier;
  short diff_level;
  // world_position3d
  int distancex; int distancey; int distancez;
  short vectori; short vectorj; short vectork; short height;
  short yaw, pitch;
  
  char user_data[16];
  short unused[4];
  short data_index, data_identifier;
};


char *tfl_data, *m2_data;
struct tfl_mesh_header tfl_header;
struct m2_mesh_header m2_header, new_header;
int m2_data_size, tfl_data_size;

int 
load_tfl(FILE *fp)
{
  int i;
  char *p;
  struct marker_data *marker;
  
  fseek(fp, 0, SEEK_END);
  i = tfl_data_size = ftell(fp);
  rewind(fp);

  fprintf(stderr, "i = %d\n",i);
  if (!(tfl_data = (char *) malloc(i)))
    return 0;

  fread(tfl_data, 1, i, fp);
  fclose(fp);
  memcpy((void *) &tfl_header, tfl_data + SKIP_SIZE, 1024);

  fprintf(stderr,"%c%c%c%c\n",(char) (tfl_header.landscape_collection_tag>>24), (char) (tfl_header.landscape_collection_tag>>16), (char) (tfl_header.landscape_collection_tag>>8), (char) (tfl_header.landscape_collection_tag));
  fprintf(stderr,"markers = %d, offset = %d, size = %d\n",tfl_header.marker_count, tfl_header.markers_offset, tfl_header.markers_size);
  fprintf(stderr,"map actions = %d, offset = %d, size = %d\n\n",tfl_header.map_action_count, tfl_header.map_actions_offset, tfl_header.map_action_buffer_size);
  
  p = (tfl_data + tfl_header.markers_offset + 1088);
  marker = (struct marker_data *) p;
  for (i=0; i<tfl_header.marker_count; i++)
  {
//fprintf(stderr,"%d: %d (%d)\n",i,marker[i].type,sizeof(struct marker_data));
    if (marker[i].type >= 12)
    {
      tfl_header.marker_count = i;
      tfl_header.markers_size = i * 64;
      fprintf(stderr,"reduced marker count to %d, size %d\n\n",tfl_header.marker_count,tfl_header.markers_size);
      break;
    }
  }
  return 1;
}

int 
load_m2(FILE *fp)
{
  int i;
  fseek(fp, 0, SEEK_END);
  i = m2_data_size = ftell(fp);
  rewind(fp);

  fprintf(stderr, "i = %d\n",i);
  if (!(m2_data = (char *) malloc(i)))
    return 0;

  fread(m2_data, 1, i, fp);
  fclose(fp);
  memcpy((void *) &m2_header, m2_data + SKIP_SIZE, 1024);

  fprintf(stderr,"%c%c%c%c\n",(char) (m2_header.landscape_collection_tag>>24), (char) (m2_header.landscape_collection_tag>>16), (char) (m2_header.landscape_collection_tag>>8), (char) (m2_header.landscape_collection_tag));
  fprintf(stderr,"markers = %d, offset = %d, size = %d\n",m2_header.marker_count, m2_header.markers_offset, m2_header.markers_size);
  fprintf(stderr,"map actions = %d, offset = %d, size = %d\n",m2_header.map_action_count, m2_header.map_actions_offset, m2_header.map_action_buffer_size);
  fprintf(stderr,"LOD offset = %d, connectors offset = %d\n\n",m2_header.mesh_LOD_data_offset,m2_header.connectors_offset);
  return 1;
}


int 
test_new(FILE *fp)
{
  int i;
  fseek(fp, 0, SEEK_END);
  i = m2_data_size = ftell(fp);
  rewind(fp);

  fprintf(stderr, "i = %d\n",i);
  if (!(m2_data = (char *) malloc(i)))
    return 0;

  fread(m2_data, 1, i, fp);
  fclose(fp);
  memcpy((void *) &m2_header, m2_data + SKIP_SIZE, 1024);

  fprintf(stderr,"%c%c%c%c\n",(char) (m2_header.landscape_collection_tag>>24), (char) (m2_header.landscape_collection_tag>>16), (char) (m2_header.landscape_collection_tag>>8), (char) (m2_header.landscape_collection_tag));
  fprintf(stderr,"markers = %d, offset = %d, size = %d\n",m2_header.marker_count, m2_header.markers_offset, m2_header.markers_size);
  fprintf(stderr,"map actions = %d, offset = %d, size = %d\n",m2_header.map_action_count, m2_header.map_actions_offset, m2_header.map_action_buffer_size);
  fprintf(stderr,"LOD offset = %d, connectors offset = %d\n\n",m2_header.mesh_LOD_data_offset,m2_header.connectors_offset);
  return 1;
}


struct map_action
{
  short identifier;
  short deact_mode;
  int tag;
  unsigned long flags;
  long trigger_low, trigger_delta;
  short param_count;
  short param_size;
  long param_offset;
  short visual_indent;
  short unused[15];
  long trigger_time;
};

struct map_action_param
{
  short type, count;
  unsigned long field_name;
};

void
save_new(FILE *fp)
{
  int offset = 0, i;
  struct map_action *map_action = (struct map_action *) (tfl_data + tfl_header.map_actions_offset);
  struct marker_data *marker = (struct marker_data *) (tfl_data + tfl_header.markers_offset);
  
  memcpy(&new_header, &m2_header, 1024);
  
  /* */
  new_header.marker_count = tfl_header.marker_count;
  new_header.markers_size = tfl_header.markers_size;
  new_header.data_size += tfl_header.markers_size;
  new_header.map_actions_offset += new_header.markers_size;
  if (new_header.data_offset >= new_header.markers_offset)
    new_header.data_offset += tfl_header.markers_size;
  if (new_header.marker_palette_offset >= new_header.markers_offset)
    new_header.marker_palette_offset += tfl_header.markers_size;
  if (new_header.media_coverage_region_offset >= new_header.markers_offset)
    new_header.media_coverage_region_offset += tfl_header.markers_size;
  if (new_header.mesh_LOD_data_offset >= new_header.markers_offset)
    new_header.mesh_LOD_data_offset += tfl_header.markers_size;
  if (new_header.connectors_offset >= new_header.markers_offset)
    new_header.connectors_offset += tfl_header.markers_size;
  if (new_header.mesh_offset >= new_header.markers_offset)
    new_header.mesh_offset += tfl_header.markers_size;
   /* */
/**/
  new_header.map_action_count = tfl_header.map_action_count;
  new_header.map_action_buffer_size = tfl_header.map_action_buffer_size;
  new_header.data_size += tfl_header.map_action_buffer_size;
  if (new_header.data_offset >= new_header.map_actions_offset)
    new_header.data_offset += tfl_header.map_action_buffer_size;
  if (new_header.marker_palette_offset >= new_header.map_actions_offset)
    new_header.marker_palette_offset += tfl_header.map_action_buffer_size;
  if (new_header.media_coverage_region_offset >= new_header.map_actions_offset)
    new_header.media_coverage_region_offset += tfl_header.map_action_buffer_size; 
  if (new_header.mesh_LOD_data_offset >= new_header.map_actions_offset)
    new_header.mesh_LOD_data_offset += tfl_header.map_action_buffer_size;
  if (new_header.connectors_offset >= new_header.map_actions_offset)
    new_header.connectors_offset += tfl_header.map_action_buffer_size;
  if (new_header.mesh_offset >= new_header.map_actions_offset)
    new_header.mesh_offset += tfl_header.map_action_buffer_size;
/**/

  // fix model positions
  //for (i=0; i<tfl_header.marker_count; i++)
  //{
   // if (marker[i].type == _marker_model)
    //{
     // marker[i].distancex += 2048;
      //marker[i].distancey += 2048;
    //}
  //}
  
  // fix wayward wayp parameters
  offset = tfl_header.map_actions_offset + 1088 + (tfl_header.map_action_count * 64);
  {
    struct map_action_param *param = (struct map_action_param *) (tfl_data + tfl_header.map_actions_offset + 1088 + (tfl_header.map_action_count * 64));
    int base_offset = offset;
    int max_offset = tfl_header.map_actions_offset + 1088 + tfl_header.map_action_buffer_size;
//fprintf(stderr,"offset %d/%d, %2x %2x %2x %2x\n",offset,max_offset,*(tfl_data + offset),*(tfl_data+offset+1),*(tfl_data+offset+2),*(tfl_data+offset+3));
    
    while (offset < max_offset)
    {
      if (param->type > 12 && param->type < 20)
      {
		param->type++;
		offset += sizeof(struct map_action_param);
		if (param->type < 10)
          offset += param->count * 4;
      }
      else if (param->type == 12 && param->count >= 0 &&
         (param->field_name == 'wayp' ||
          param->field_name == 'cent' ||
          param->field_name == 'deci' ||
          param->field_name == 'line' ||
          param->field_name == 'grce' ||
          param->field_name == 'poly' ||
          param->field_name == 'poin' ||
          param->field_name == 'focu' ||
          param->field_name == 'dest' ||
          param->field_name == 'pnts' ||
          param->field_name == 'radi'))
      {
//       param->count *= 2;
		param->type++;
        offset += sizeof(struct map_action_param) + param->count * 4;
      }
      else if (param->type >= 12 && param->type < 20 && param->count >= 0)
      {
        fprintf(stderr,"%c%c%c%c\n",(char) (param->field_name >> 24), (char) (param->field_name >> 16), (char) (param->field_name >> 8), (char) param->field_name);
        offset += 4;
      }
      else
      {
        offset += 4;
      }
      param = (struct map_action_param *) (tfl_data + offset);
    }
  }
  
  // save the global header
  fwrite(m2_data, 1, SKIP_SIZE, fp);
  
  // save the new mesh header
  fwrite(&new_header, 1, 1024, fp);
  
  // save the data up to the markers offset
  offset = SKIP_SIZE + 1024;
  fwrite(m2_data + offset, 1, new_header.markers_offset, fp);
//fprintf(stderr,"from headers to markers offset: %d to %d\n",offset,offset + (new_header.markers_offset - offset));
  
  // insert TFL marker data
  fwrite(tfl_data + tfl_header.markers_offset + 1088, 1, tfl_header.markers_size, fp);
  //offset = new_header.markers_size + new_header.markers_offset;
//fprintf(stderr,"tfl markers: %d to %d (%d)\n",tfl_header.markers_offset, tfl_header.markers_offset + tfl_header.markers_size, tfl_data_size);
    
  // save the TFL map actions
  fwrite(tfl_data + tfl_header.map_actions_offset + 1088, 1, tfl_header.map_action_buffer_size, fp);
//fprintf(stderr,"tfl map actionss: %d to %d (%d)\n",tfl_header.map_actions_offset, tfl_header.map_actions_offset + tfl_header.map_action_buffer_size, tfl_data_size);
  
  // write the remainder of the file
  offset = new_header.markers_offset;
  fwrite(m2_data + offset, 1, m2_data_size - offset, fp);
//fprintf(stderr,"from map actions to end (%d): %d to %d\n\n",m2_data_size,offset,offset + (m2_data_size - offset));
  fclose(fp);
}

#define MAXPATHL 2048

int main()
{
  static char m2file[1024], tflfile[1024], newfile[1024];
  FILE *tflfp, *m2fp, *newfp;
  char c;

  MaxApplZone();
  SIOUXSettings.asktosaveonclose = FALSE;
  SIOUXSettings.rows = 15;
  SIOUXSettings.toppixel = 46;
  SIOUXSettings.leftpixel = 8;
  SIOUXSettings.autocloseonquit = TRUE;
  SIOUXSetTitle("\pTFL to M2 Mesh Converter");  
  clrscr();
  
  if (!(tflfp = fopen("tflmesh", "rb")))
  {
    fprintf(stderr,"Couldn't find the file 'tflmesh'\nHit any key to quit\n");
    c = getch();
    exit(1);
  }
  fprintf(stderr,"Loading tflmesh...\n");

  if (!load_tfl(tflfp))
  {
    exit(1);
  }
  
  if (!(m2fp = fopen("m2mesh", "rb")))
  {
    fprintf(stderr,"Couldn't find the file 'm2mesh'\nHit any key to quit\n");
    c = getch();
    exit(1);
  }
  fprintf(stderr,"Loading m2mesh...\n");

  if (!load_m2(m2fp))
  {
    exit(1);
  }
  
  if (!(newfp = fopen("newmesh", "wb")))
  {
    fprintf(stderr,"Couldn't find the file 'newmesh'\nHit any key to quit\n");
    c = getch();
    exit(1);
  }
  save_new(newfp);
  
  if (!(newfp = fopen("newmesh", "rb")))
  {
    fprintf(stderr,"Couldn't find the file 'newmesh'\nHit any key to quit\n");
    c = getch();
    exit(1);
  }
  test_new(newfp);

  fprintf(stderr, "Done\n");
  
  c = getch();
}

